#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <itdb.h>

/*
  itdb2tc - iTunes DB to Rockbox Tag Cache converter 0.1
  by Adam Gashlin (hcs)
  with stuff from songdb.pl, iTunesDB reading by libgpod
*/

gint compareArtist(gconstpointer gcpa, gconstpointer gcpb) {
    Itdb_Track const * a = gcpa, * b = gcpb;
    return strcmp(
        (a->artist?a->artist:""),
        (b->artist?b->artist:""));
}

gint compareAlbum(gconstpointer gcpa, gconstpointer gcpb) {
    Itdb_Track const * a = gcpa, * b = gcpb;
    return strcmp(
        (a->album?a->album:""),
        (b->album?b->album:""));
}

gint compareGenre(gconstpointer gcpa, gconstpointer gcpb) {
    Itdb_Track const * a = gcpa, * b = gcpb;
    return strcmp(
        (a->genre?a->genre:""),
        (b->genre?b->genre:""));
}

gint compareTitle(gconstpointer gcpa, gconstpointer gcpb) {
    Itdb_Track const * a = gcpa, * b = gcpb;
    return strcmp(
        (a->title?a->title:""),
        (b->title?b->title:""));
}

gint comparePath(gconstpointer gcpa, gconstpointer gcpb) {
    Itdb_Track const * a = gcpa, * b = gcpb;
    return strcmp(
        (a->ipod_path?a->ipod_path:""),
        (b->ipod_path?b->ipod_path:""));
}

gint compareComposer(gconstpointer gcpa, gconstpointer gcpb) {
    Itdb_Track const * a = gcpa, * b = gcpb;
    return strcmp(
        (a->composer?a->composer:""),
        (b->composer?b->composer:""));
}

gint compareTracks(gconstpointer gcpa, gconstpointer gcpb) {
    Itdb_Track const * a = gcpa, * b = gcpb;

    if (a->cd_nr-b->cd_nr) return a->cd_nr-b->cd_nr;
    if (a->track_nr-b->track_nr) return a->track_nr-b->track_nr;

    return strcmp(a->title,b->title);
}

enum keytype {KEY_ARTIST=0,KEY_ALBUM,KEY_GENRE,KEY_TITLE,KEY_PATH,KEY_COMPOSER,KEY_COUNT};

gint (*compareKey[KEY_COUNT])(gconstpointer, gconstpointer) = {compareArtist,compareAlbum,compareGenre,compareTitle,
    comparePath,compareComposer};

struct tagcache_meta {
        int keyoffs[KEY_COUNT];
        int index;
};


    
char * getKey(Itdb_Track * track, enum keytype key) {
    switch (key) {
    case KEY_ARTIST:
        return (track->artist?track->artist:"");
    case KEY_ALBUM:
        return (track->album?track->album:"");
    case KEY_GENRE:
        return (track->genre?track->genre:"");
    case KEY_TITLE:
        return (track->title?track->title:" ");
    case KEY_PATH:
        {
            int i;
            if (!(track->ipod_path)) return "";
            for (i=0;track->ipod_path[i];i++) {
    	   	    if (track->ipod_path[i]==':') track->ipod_path[i]='/';
	       	}
	    }
        return track->ipod_path;
    case KEY_COMPOSER:
        return (track->composer?track->composer:"");
    default:
        return NULL;
    }
}

/* 1 if successful */
size_t dumpshort(FILE * db, unsigned short num, int littleendian) {
    unsigned char buf[2];
    
    if (littleendian) {
        buf[0]= num&0x00ff;
        buf[1]=(num&0xff00)>>8;
    } else {    
        buf[0]=(num&0xff00)>>8;
        buf[1]= num&0x00ff;
    }
    
    return fwrite(buf,2,1,db);
}

/* 1 if successful */
size_t dumpint(FILE * db, unsigned long num, int littleendian) {
    unsigned char buf[4];
    
    if (littleendian) {
        buf[0]= num&0x000000ff;
        buf[1]=(num&0x0000ff00)>>8;
        buf[2]=(num&0x00ff0000)>>16;
        buf[3]=(num&0xff000000)>>24;
    } else {    
        buf[0]=(num&0xff000000)>>24;
        buf[1]=(num&0x00ff0000)>>16;
        buf[2]=(num&0x0000ff00)>>8;
        buf[3]= num&0x000000ff;
    }
    
    return fwrite(buf,4,1,db);
}

/* # of bytes written if successful */
size_t dump_tag_string(FILE * db, char * s, unsigned long index, int littleendian) {
    int len = strlen(s)+1;
    int padding = len%4;
    size_t r;
    char zeros[4]={0,0,0,0};
    int i;
    
    if (padding > 0) {
        padding = 4 - padding;
        len += padding;
    }

    if ((r=dumpshort(db,len,littleendian)) != 1) return r;
    if ((r=dumpshort(db,index,littleendian)) != 1) return r;
    if ((r=fwrite(s,1,len-padding,db)) != len-padding) return r;
    if ((r=fwrite(zeros,1,padding,db)) != padding) return r;
    
    return 2+2+len+padding;
}

/* # of bytes written if successful */
size_t dump_tag_header(FILE * db, unsigned long dbver, unsigned long entry_count, int littleendian) {
    size_t size;
    
    size = fseek(db, 0, SEEK_END) - 12;
    fseek(db, 0, SEEK_SET);
    
    dumpint(db,dbver,littleendian);
    dumpint(db,size,littleendian);
    dumpint(db,entry_count,littleendian);
}

/* 0 on failure */

int create_tagcache_index_file(GList * itdb, char * prefix, unsigned int index, enum keytype key, unsigned long dbver, int unique, int littleendian) {
    int num = 0;
    char dbname[0x101] = "";
    long offset = 12;
    FILE * db = NULL;
    GList * sorted = NULL, * iterator = NULL;

    sprintf(dbname,"%s_%d.tcd",prefix,index);
    db = fopen(dbname,"wb");

    fseek(db, offset, SEEK_SET);    
    
    sorted = g_list_sort(g_list_copy(itdb), compareKey[key]);
    iterator = g_list_first(sorted);
    
    while (iterator) {
        struct tagcache_meta * my_meta = (((Itdb_Track *)(iterator->data))->userdata);
        if (!unique || !(iterator->prev) || (strcmp(getKey(iterator->prev->data,key),getKey(iterator->data,key)))) {
            num++;
            offset = ftell(db);
            //printf("  %s\n",getKey(iterator->data,key));
            
            if (unique) index = -1;
            else index = my_meta->index;
            dump_tag_string(db,getKey(iterator->data,key),index,littleendian);
        }
        my_meta->keyoffs[key] = offset;
        iterator=g_list_next(iterator);
    }

    dump_tag_header(db,dbver,num,littleendian);
    
    fclose(db); db=NULL;
    
    g_list_free(sorted); sorted=NULL;
}

int main(int argc, char ** argv) {
    GError * error = NULL;
    Itdb_iTunesDB * db = NULL;
    GList * iterator = NULL, *sorted = NULL;
    FILE * master;
    int i=0;
    int littleendian = 1;
    char prefix[0x101] = "tagcache";
    char outpath[0x121];
    unsigned long dbver = 0x54434806;

    fprintf(stderr,"itdb2tc - iTunes DB to Rockbox Tag Cache (little endian) converter 0.1\n\n");

    if (argc<2) {
	fprintf(stderr,"usage: %s iTunesDB\n",argv[0]);
	return 1;
    }
    
    printf("Loading iTunesDB\n");

    db = itdb_parse_file(argv[1],&error);

    if (!db) {
	fprintf(stderr,"error opening %s\n",argv[1]);
	return 1;
    }
    
    /* index each track */
    printf("Indexing\n");
    sorted = g_list_sort(g_list_copy(db->tracks),comparePath);
    iterator = g_list_first(sorted);
    for (i=0;iterator;iterator = g_list_next(iterator),i++) {
        Itdb_Track * track = iterator->data;

        track->userdata = malloc(sizeof(struct tagcache_meta));
        if (!track->userdata) {
        fprintf(stderr,"out of memory\n");
        return 1;
        }
        ((struct tagcache_meta*)(track->userdata))->index=i;
    }
    
    /* tagcache index files */
    
    printf("Building table of Artists\n");
    create_tagcache_index_file(db->tracks,prefix,0,KEY_ARTIST,dbver,1,littleendian);
    printf("Building table of Albums\n");
    create_tagcache_index_file(db->tracks,prefix,1,KEY_ALBUM,dbver,1,littleendian);
    printf("Building table of Genres\n");
    create_tagcache_index_file(db->tracks,prefix,2,KEY_GENRE,dbver,1,littleendian);
    printf("Building table of Titles\n");
    create_tagcache_index_file(db->tracks,prefix,3,KEY_TITLE,dbver,0,littleendian);
    printf("Building table of Paths\n");
    create_tagcache_index_file(db->tracks,prefix,4,KEY_PATH,dbver,0,littleendian);
    printf("Building table of Composers\n");
    create_tagcache_index_file(db->tracks,prefix,5,KEY_COMPOSER,dbver,1,littleendian);
    
    /* Master index file */
    printf("Building master index\n");
    sprintf(outpath,"%s_idx.tcd",prefix);
    master = fopen(outpath,"wb");
    if (!master) {
    fprintf(stderr,"error opening %s for output\n",outpath);
    return 1;
    }
    fseek(master,12,SEEK_SET);

    /* current serial */
    dumpint(master,0,littleendian);
    
    iterator = g_list_first(sorted);
    while (iterator) {
        Itdb_Track * my_track = iterator->data;
        struct tagcache_meta * my_meta = my_track->userdata;
        
        dumpint(master,my_meta->keyoffs[KEY_ARTIST],littleendian);
        dumpint(master,my_meta->keyoffs[KEY_ALBUM],littleendian);
        dumpint(master,my_meta->keyoffs[KEY_GENRE],littleendian);
        dumpint(master,my_meta->keyoffs[KEY_TITLE],littleendian);
        dumpint(master,my_meta->keyoffs[KEY_PATH],littleendian);
        dumpint(master,my_meta->keyoffs[KEY_COMPOSER],littleendian);
        dumpint(master,my_track->year,littleendian);
        dumpint(master,my_track->track_nr+(my_track->cd_nr)*100,littleendian);
        dumpint(master,my_track->bitrate,littleendian);
        
        /* track length, play count, play time */
        if (my_track->rating > 0) {
            int length, playcount, playtime;
            /*
              Rating is not stored in the tagcache, it is
              calculated as follows:
            
                rating = 100 * playtime / length / playcount

              thus we can fudge the playtime to fake a rating:
              
                playtime = rating * length * playcount / 100
              
              It uses the same scale (0-100) that iTunes does.
              
              This also means that we must fake a play count or length
              if the track does not have one.
            */
            if ((length=my_track->tracklen/1000)==0)
                length=1;
            dumpint(master,length,littleendian); /* length */
            if ((playcount=my_track->playcount)==0)
                playcount=1;
            dumpint(master,playcount,littleendian); /* play count */
            
            playtime = my_track->rating * length * playcount / 100;
            dumpint(master,playtime,littleendian); /* play time */
        } else {
            dumpint(master,my_track->tracklen,littleendian); /* length */
            dumpint(master,my_track->playcount,littleendian); /* play count */
            dumpint(master,0,littleendian); /* play time */
        }
        
        
        /* last played (not sure how to parse the date in iTunesDB yet) */
        dumpint(master,0,littleendian);
        /* status flag */
        dumpint(master,0,littleendian);
        
        if (my_meta) free(my_meta);
        my_track->userdata=NULL;
        
        iterator=g_list_next(iterator);
    }
    
    g_list_free(sorted); sorted=NULL;

    dump_tag_header(master,dbver,i,littleendian);
    fclose(master); master=NULL;
    
    itdb_free(db); db=NULL;
    return 0;
}
